#include <Wire.h>
#include <EEPROM.h>

#include "DHT.h"
#include "LCD16x2.h"

#define DHTPIN 2
#define DHTTYPE DHT22

#define RELAYPIN 3

#define TMIN 45
#define TMAX 90
#define TOVERSHOOT 2

#define INFO1 "Dual Set Point  "
#define INFO2 "Thermostat V1.2a"  

#define VALID_EEPROM_HEADER "TSTAT EEPROM V1.0"

DHT dht(DHTPIN, DHTTYPE);

LCD16x2 lcd;

int TempSet;
int TempAct;
int RH;

int T1;
int T2;
int TX;

int Heat;

unsigned long STimer=0;
unsigned long UTimer=0;
unsigned long BTimer=0;
unsigned long HTimer=0;

char Line1[17];
char Line2[17];

void setup(){
  
  delay(1000);
  
  Wire.begin();
  dht.begin();
  
  pinMode(RELAYPIN, OUTPUT);
  
  if(EEPROMVerifyHeader(VALID_EEPROM_HEADER)) {
    EEPROMGetData();
  } else {
    EEPROMSetHeader(VALID_EEPROM_HEADER);
    TX=1;
    T1=TMIN;
    T2=TMAX;   
    EEPROMSetData();
  }
  
  if(TX == 1) TempSet = T1;
  else TempSet = T2;
  
  Heat=0;

  lcd.lcdClear();
  lcd.lcdSetBlacklight(128);
  DisplayWelcome();  
  delay(5000);

  ReadSensor();
  STimer=SetTimer();                                   //  Start the Sensor Update Timer

  lcd.lcdClear();
  UpdateDisplay();
  BTimer=SetTimer();                                   //  Start the Backlight Timer  
}

void loop(){
int update;
int buttons;

  update=0;

  buttons = lcd.readButtons();

  if (!(buttons & 0x01)) {                               //  "T1" Button Pressed
    TX=1;
    update++;
  }

  if (!(buttons & 0x02)) {                               // "-" Button Pressed
    if(TX == 1) T1 = DecreaseTemp(T1);
    else T2 = DecreaseTemp(T2);
    update++;
  }

  if (!(buttons & 0x04)) {                               // "+" Button Pressed
    if(TX == 1) T1 = IncreaseTemp(T1);
    else T2 = IncreaseTemp(T2);
    update++;
  }

  if (!(buttons & 0x08)) {                               // "T2" Button Pressed
    TX=2;
    update++;
  }
  
  if(TX == 1) TempSet = T1;
  else TempSet = T2;

  if(update) {                                           //  Somebody Pressed a Button, so...
    UpdateDisplay();                                     //  Update the Display
    lcd.lcdSetBlacklight(128);                           //  Turn On the Backlight
    BTimer=SetTimer();                                   //  Start the Backlight Timer
    UTimer=SetTimer();                                   //  Start the Update Timer
  } 
  
  if(CheckTimer(STimer, 2.5 * 1000)) {                   //  Check the Sensor Timer (2.5 seconds)
      ReadSensor();                                      //  Get new sensor readings,
      UpdateDisplay();                                   //  Update the Display,
      STimer=SetTimer();                                 //  and RESTART the Timer 
  }
  
  if(CheckTimer(BTimer, 10 * 1000)) {                    //  Check the Backlight Timer (10 Seconds)
      lcd.lcdSetBlacklight(0);                           //  Turn Off The Backlight,
      BTimer=0;                                          //  and CLEAR the Timer 
  }

  if(CheckTimer(UTimer, 60 * 1000)) {                    //  Check the EEPROM Update Timer (60 Seconds)
      EEPROMSetData();                                   //  Save new setpoints to EEPROM
      UTimer=0;                                          //  and CLEAR the Timer 
  }
  
  HeatControl();  
}
 
int DecreaseTemp(int value) {
  if(value == 255) return TMAX;
  else if(value > TMIN) return (value-1);
  else return 0;
}

int IncreaseTemp(int value) {
  if(value == 0) return TMIN;
  else if(value < TMAX) return (value+1);
  else return 255;
}

void HeatControl() {
  
  if(TempAct == 255) {                                    // Heat Off if Sensor Failure
    HeatOff();
    return;
  }
    
  if(TempSet == 0) HeatOff();
  else if(TempSet == 255) HeatOn();
  else if(TempAct < TempSet) HeatOn();
  else if(TempAct > TempSet + TOVERSHOOT) HeatOff();
}

void HeatOn() {
    if(Heat == 0) {                                       // Heat is OFF
      digitalWrite(RELAYPIN, HIGH);
      Heat = 1;
      lcd.lcdGoToXY(6,2);
      lcd.lcdWrite("*");

      HTimer=SetTimer();                                  // Set Max Heating Cycle Timer (Detects tripped Heater Circuit)

    } else {                                              // Heat is already on

      if(CheckTimer(HTimer, 60 * 60 * 1000)) {            // If the Heat has been on constantly (1 Hr)
        digitalWrite(RELAYPIN, LOW);                      // Turn it off
        delay(10 * 1000);                                 // Wait 10 seconds (to reset the heater)
        digitalWrite(RELAYPIN, HIGH);                     // Then Turn it back on again
        HTimer=SetTimer();                                // and restart the timer
      }
  
    }
}

void HeatOff() {
    if(Heat == 1) {
      digitalWrite(RELAYPIN, LOW);
      Heat = 0;
      lcd.lcdGoToXY(6,2);
      lcd.lcdWrite(" ");
      
      HTimer=0;                                          // Clear the Max Heating Cycle Timer
    }
}

void UpdateDisplay(){
char HeatChar;

  if(TempAct == 255) {
    lcd.lcdSetBlacklight(128);
    strcpy(Line1," *** ERROR ***  ");
    strcpy(Line2," Sensor Failure ");
  } else {

    if(Heat == 1) HeatChar='*';
    else HeatChar=' ';

    sprintf(Line1,"Set(T%d)  Actual ",TX);

    if(TempSet == 0) sprintf(Line2," OFF %c %3dF %3d%%",HeatChar,(int)TempAct,(int)RH);
    else if(TempSet == 255) sprintf(Line2," ON  %c %3dF %3d%%",HeatChar,(int)TempAct,(int)RH);
    else sprintf(Line2,"%3dF %c %3dF %3d%%",TempSet,HeatChar,(int)TempAct,(int)RH);
  }

  lcd.lcdGoToXY(1,1);
  lcd.lcdWrite(Line1);
  lcd.lcdGoToXY(1,2);
  lcd.lcdWrite(Line2);

  delay(100);
}

void ReadSensor() {
float h;
float f;

  h = dht.readHumidity();
  f = dht.readTemperature(true);
  
  if(((int)f == 32) && ((int)h == 0)) TempAct=255;   // 32F & 0%RH = Flag Sensor Comm Failure
  else {
    TempAct=(int)f;
    RH=(int)h;
  }
}

void DisplayWelcome(){
  lcd.lcdGoToXY(1,1);
  lcd.lcdWrite(INFO1);
  lcd.lcdGoToXY(1,2);
  lcd.lcdWrite(INFO2);  
}

unsigned long SetTimer() {
unsigned long tnow;

  tnow=millis();  
  if(tnow == 0) tnow=1;           // What are the odds?  Well... 
  return tnow;
}

int CheckTimer(unsigned long timer, unsigned long ms) {
unsigned long tnow;
unsigned long ttarget;

  if(timer == 0) return 0;

  tnow=millis();
  ttarget=timer+ms;  

  if(tnow > ttarget) {
    if(ttarget < timer) {
      if(tnow < timer) return 1;
    } else return 1;
  }

  return 0;
}

int EEPROMVerifyHeader(char *header) {
int i;
int len;

  len=strlen(header);
  for(i=0; i<len; i++) {
    if(header[i] != EEPROM.read(i)) return 0;
  }
  return 1;
}

void EEPROMSetHeader(char *header) {
int i;
int len;

  len=strlen(header);
  for(i=0; i<len; i++) EEPROM.write(i,header[i]);
}

void EEPROMGetData() {
  TX=EEPROM.read(20);
  T1=EEPROM.read(21);
  T2=EEPROM.read(22);
}

void EEPROMSetData() {
  EEPROM.write(20,(byte)TX);
  EEPROM.write(21,(byte)T1);
  EEPROM.write(22,(byte)T2);
}

